import os

import tensorflow as tf
import numpy as np
import tifffile

from ..util.gs import ensure_local_file


def get_mask(masktype, h, w, job_dir='.'):
    if masktype[-4:] == '.pad':
        pad = int(masktype.split('.')[-2])
        mask = np.zeros([h, w], bool)
        mask[pad:h-pad, pad:w-pad] = True
    else:
        if masktype[:2] == 'r:':
            maskfile = ensure_local_file(os.path.join(job_dir, masktype[2:]))
        else:
            maskfile = ensure_local_file(os.path.join(job_dir, masktype))
        if masktype[-4:] == '.tif':
            mask = tifffile.imread(maskfile)
        elif masktype[-4:] == '.npy':
            mask = np.load(maskfile)
        else:
            raise RuntimeErorr('bad file type: {maskfile}')
        mask = mask > 0
        if masktype[:2] == 'r:':
            mask = ~mask
    return mask


def get_mask_range(mask):
    my = np.where(np.any(mask, axis=1))[0]
    mx = np.where(np.any(mask, axis=0))[0]
    return my[0], my[-1] + 1, mx[0], mx[-1] + 1



import tensorflow.keras.backend as K
import tensorflow as tf
from tqdm import trange

from ..util.distribute import distributed, ReduceOp
from ..util.dataset import unmasked
from ..image.filter.gaussian import gaussian
from ..image.filter.laplace import gaussian_laplace_multi
from .segment import get_segment_index
from .util import get_normalized_val, get_magnitude, ToDense


def make_segment(dataset, mask, gauss, radius, peaks, shard, batch, verbose):
    ts, rs, ys, xs = peaks[:, 0], peaks[:, 1], peaks[:, 2], peaks[:, 3]
    nk = ts.size

    dataset = unmasked(dataset.shard(shard, 0), mask)
    dataset = dataset.batch(batch)
    to_dense = ToDense(mask)

    gauss = K.constant(gauss, tf.float32)
    radius = K.constant(radius, tf.float32)
    ts = K.constant(ts / shard, tf.int32)
    rs = K.constant(rs, tf.int32)
    ys = K.constant(ys, tf.int32)
    xs = K.constant(xs, tf.int32)

    fps = tf.TensorArray(tf.float32, 0, True)
    e = K.constant(0, tf.int32)
    with trange(nk, disable=verbose==0) as prog:
        for imgs in dataset:
            s, e = e, e + tf.shape(imgs)[0]
            gl, yl, xl = _prepare(
                imgs, s, e, ts, rs, ys, xs, gauss, radius,
            )
            nk = tf.size(yl)
            for k in tf.range(nk):
                g, y, x = gl[k], yl[k], xl[k]
                pos = get_segment_index(g, y, x, mask)
                val = get_normalized_val(g, pos)
                if get_magnitude(g, pos) > 0.0:
                    footprint = to_dense(pos, val)
                    fps = fps.write(fps.size(), footprint)
                prog.update(1)
    return fps.stack().numpy()


@distributed(*[ReduceOp.CONCAT for _ in range(5)], loop=False)
def _prepare(imgs, start, end, ts, rs, ys, xs, gauss, radius):
    cond = (start <= ts) & (ts < end)
    ids = tf.cast(tf.where(cond)[:, 0], tf.int32)
    tl = tf.gather(ts, ids) - start
    rl = tf.gather(rs, ids)
    yl = tf.gather(ys, ids)
    xl = tf.gather(xs, ids)
    if gauss > 0.0:
        imgs = gaussian(imgs, gauss)
    gls = gaussian_laplace_multi(imgs, radius)
    gl = tf.gather_nd(gls, tf.stack([tl, rl], 1))
    return gl, yl, xl



import tensorflow as tf
import pickle


def load_pickle(base):
    with tf.io.gfile.GFile(f'{base}.pickle', 'rb') as fp:
        return pickle.load(fp)


def save_pickle(base, val):
    with tf.io.gfile.GFile(f'{base}.pickle', 'wb') as fp:
        pickle.dump(val, fp)



import tensorflow.keras.backend as K
import tensorflow as tf
import numpy as np

from .regularizer import ProxOp, MaxNormNonNegativeL1


class DynamicInputLayer(tf.keras.layers.Layer):

    def __init__(self, nk, nx, regularizer=None, name='input', **kwargs):
        super().__init__(name=f'{name}_layer', **kwargs)
        regularizer = regularizer or ProxOp()
        self.max_nk = nk
        self._nk = self.add_weight(f'{name}/nk', (), tf.int32, trainable=False)
        self._val = self.add_weight(f'{name}/val', (nk, nx))
        self._val.regularizer = regularizer

    @property
    def l(self):
        return K.get_value(self._val.regularizer.l)

    @l.setter
    def l(self, val):
        self.set_l(l)

    @property
    def val(self):
        return self.get_val()

    @val.setter
    def val(self, val):
        self.set_val(val)

    def set_l(self, val):
        K.set_value(self._val.regularizer.l, val)

    def get_l(self):
        return K.get_value(self._val.regularizer.l)

    def set_val(self, val):
        nk = val.shape[0]
        val = np.pad(val, ((0, self.max_nk - nk), (0, 0)))
        K.set_value(self._nk, nk)
        K.set_value(self._val, val)

    def get_val(self):
        return K.get_value(self.call())

    def penalty(self, x=None):
        if x is None:
            x = self.call()
        return self._val.regularizer(x)

    def call(self, dummy=None):
        nk = self._nk
        nx = tf.shape(self._val)[1]
        return tf.slice(self._val, (0, 0), (nk, nx))


class MaxNormNonNegativeL1InputLayer(DynamicInputLayer):

    def __init__(self, nk, nx, name='mnnnl1', **kwargs):
        reg = MaxNormNonNegativeL1(1)
        super().__init__(nk, nx, reg, name=name, **kwargs)



from enum import Enum

import tensorflow as tf


class ReduceOp(Enum):
    CONCAT = 0
    SUM = 1
    MAX = 2


def distributed(*types, loop=True):

    def make_init(t):
        if t == ReduceOp.SUM:
            return tf.constant(0, tf.float32)
        elif t == ReduceOp.MAX:
            return ()
        elif t == ReduceOp.CONCAT:
            return ()
        else:
            raise NotImplementedError()

    def serialize(x, t):
        if t == ReduceOp.SUM:
            return strategy.reduce(tf.distribute.ReduceOp.SUM, x, axis=None)
        elif t == ReduceOp.MAX:
            return strategy.experimental_local_results(x)
        elif t == ReduceOp.CONCAT:
            return strategy.experimental_local_results(x)
        else:
            raise NotImplementedError()

    def finish(o, t):
        if t == ReduceOp.SUM:
            return o
        elif t == ReduceOp.MAX:
            return tf.reduce_max(o, axis=0)
        elif t == ReduceOp.CONCAT:
            return tf.concat(o, axis=0)
        else:
            raise NotImplementedError()

    def decorator(func):

        @tf.function(experimental_relax_shapes=True)
        def dist_run(*args, **kwargs):
            xs = strategy.run(func, args, kwargs)
            return tuple(serialize(x, t) for x, t in zip(xs, types))

        def loop_run(data, *args, prog=None, **kwargs):
            os = tuple(make_init(t) for t in types)
            for d in strategy.experimental_distribute_dataset(data):
                xs = dist_run(d, *args, **kwargs)
                os = tuple(o + x for o, x in zip(os, xs))
                if prog is not None:
                    if isinstance(d, tuple):
                        prog.update(tf.shape(d[0])[0].numpy())
                    else:
                        prog.update(tf.shape(d)[0].numpy())
            return tuple(finish(o, t) for o, t in zip(os, types))

        def single_run(*args, **kwargs):
            xs = dist_run(*args, **kwargs)
            return tuple(finish(x, t) for x, t in zip(xs, types))

        return loop_run if loop else single_run

    strategy = tf.distribute.get_strategy()
    return decorator



import tensorflow.keras.backend as K
import tensorflow as tf


def get_normalized_val(img, pos):
    val = tf.gather_nd(img, pos)
    min_val = K.min(val)
    max_val = K.max(val)
    return (val - min_val) / (max_val - min_val)


def get_magnitude(img, pos):
    val = tf.gather_nd(img, pos)
    return K.max(val) - K.min(val)


class ToDense(object):

    def __init__(self, mask):
        mask = K.constant(mask, tf.bool)
        nx = tf.cast(tf.math.count_nonzero(mask), tf.int32)
        ids = tf.cast(tf.where(mask), tf.int32)
        rmap = tf.scatter_nd(ids, tf.range(nx) + 1, tf.shape(mask)) - 1
        self.rmap = rmap
        self.nx = nx

    def __call__(self, pos, val):
        pos = tf.gather_nd(self.rmap, pos)
        out = tf.scatter_nd(pos[:, None], val, (self.nx,))
        return out



import os

import tensorflow as tf
import numpy as np

from cleo import Command as CommandBase
from cleo import option

from .status import Status
from ..util.dataset import normalized
from ..util.tfrecord import load_tfrecord
from ..util.numpy import load_numpy
from ..util.pickle import load_pickle, save_pickle

from ..optimizer.input import MaxNormNonNegativeL1InputLayer as Input
from ..train.variance import Variance


def _option(*args):
    return option(*args, flag=False, value_required=False)


class Command(CommandBase):

    def set_job_dir(self, default='.'):
        current_job_dir = self.application.job_dir
        job_dir = self.option('job-dir')
        if current_job_dir is None:
            self.application.job_dir = job_dir or default
            status_base = os.path.join(self.work_dir, 'status')
            if tf.io.gfile.exists(f'{status_base}.pickle'):
                self.application.status = load_pickle(status_base)
            else:
                tf.io.gfile.makedirs(self.work_dir)
                self.application.status = Status()
        elif job_dir and job_dir != current_job_dir:
            raise RuntimeError('config mismatch: job-dir')

    @property
    def work_dir(self):
        return os.path.join(self.application.job_dir, 'work')

    @property
    def logs_dir(self):
        return os.path.join(self.application.job_dir, 'logs')

    @property
    def status(self):
        return self.application.status

    def save_status(self):
        status_base = os.path.join(self.work_dir, 'status')
        save_pickle(status_base, self.status)

    def get_model(self, data, tau, nk):
        tfrecord = load_tfrecord(f'{data}-data')
        mask = load_numpy(f'{data}-mask')
        nx, nt = load_pickle(f'{data}-stat')[:2]
        variance = Variance(tfrecord, nk, nx, nt)
        variance.set_double_exp(*tau)
        footprint = Input(nk, nx, name='footprint')
        footprint.mask = mask
        spike = Input(nk, variance.nu, name='spike')
        return footprint, spike, variance

    def handle(self):
        self.set_job_dir()
        name = self.status.params['name']
        history = self.status.history.get(name, ())
        stage = len(history)
        if self.option('force'):
            history = history[:self.force_stage(stage)]
            stage = len(history)
        if self.is_error(stage):
            self.line(f'invalid stage: {self.name} {stage}', 'error')
            return 1
        elif self.is_target(stage):
            self.line(f'{self.name} ({stage})', 'info')
            key = self.status.get_params(stage)
            history = history + (key,)
            curr = self.status.find_saved(history)
            if self.option('force') or not curr:
                data = self.status.find_saved(history[:1])
                if data:
                    data = os.path.join(self.work_dir, data, '000')
                prev = self.status.find_saved(history[:-1])
                if prev:
                    prev = os.path.join(self.work_dir, prev, f'{stage-1:03d}')
                curr = os.path.join(self.work_dir, name, f'{stage:03d}')
                logs = os.path.join(self.logs_dir, name, f'{stage:03d}')
                tf.io.gfile.makedirs(os.path.join(self.work_dir, name))
                self.create(data, prev, curr, logs, *key)
                self.status.add_saved(history, name)
            self.status.history[name] = history
            self.save_status()
            return 0



import tensorflow as tf
import numpy as np

from .base import Command, option, _option
from ..train.footprint import FootprintModel
from ..util.numpy import load_numpy, save_numpy
from ..util.pickle import load_pickle


class FootprintCommand(Command):

    name = 'footprint'
    description = 'Update footprint'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
        option('force', 'f'),
    ]

    def is_error(self, stage):
        return stage < 4

    def is_target(self, stage):
        return stage % 3 == 1

    def force_stage(self, stage):
        return 3 * ((stage - 1) // 3) + 1

    def create(self, data, prev, curr, logs, la, bx, bt):
        tau = load_pickle(f'{prev}-tau')
        spike = load_numpy(f'{prev}-spike')
        nk = spike.shape[0]
        lu = self.status.params['lu']
        #with self.application.strategy.scope():
        elems = self.get_model(data, tau, nk)
        model = FootprintModel(*elems)
        model.compile()
        model.set_penalty(la, lu, bx, bt)
        model.fit(
            spike, stage=curr[-3:],
            lr=self.status.params['learning-rate'],
            batch=self.status.params['batch'],
            steps_per_epoch=self.status.params['step'],
            epochs=self.status.params['epoch'],
            verbose=self.status.params['pbar'],
            min_delta=self.status.params['tol'],
            log_dir=logs,
        )
        save_numpy(f'{curr}-footprint', model.footprint.val)



import tensorflow.keras.backend as K
import tensorflow as tf


def normalized(data, avgt, avgx, std):

    def normalize(img, avgt):
        return (img - avgt - avgx) / std

    avgt = tf.data.Dataset.from_tensor_slices(avgt)
    avgx = K.constant(avgx)
    std = K.constant(std)
    zip_data = tf.data.Dataset.zip((data, avgt))
    return zip_data.map(normalize)


def masked(data, mask):
    mask = K.constant(mask, tf.bool)
    return data.map(lambda x: tf.boolean_mask(x, mask))


def unmasked(data, mask):
    mask = K.constant(mask, tf.bool)
    rmap = tf.cast(tf.where(mask), tf.int32)
    shape = tf.shape(mask)
    return data.map(lambda x: tf.scatter_nd(rmap, x, shape))



import tensorflow.keras.backend as K
import tensorflow as tf
from tqdm import trange

from ..util.distribute import distributed, ReduceOp
from ..util.dataset import unmasked
from ..image.filter.gaussian import gaussian
from ..image.filter.laplace import gaussian_laplace_multi


def find_peak(
        data, mask, gauss, radius, thr_intensity, shard,
        batch, nt=None, verbose=1,
):

    @distributed(ReduceOp.CONCAT, ReduceOp.CONCAT)
    def _find(data, mask, gauss, radius, thr_intensity):
        times, imgs = data
        times = tf.cast(times, tf.int32)
        if gauss > 0.0:
            imgs = gaussian(imgs, gauss)
        gl = gaussian_laplace_multi(imgs, radius)
        max_gl = tf.nn.max_pool3d(
            gl[..., None], [1, 3, 3, 3, 1], [1, 1, 1, 1, 1], padding='SAME'
        )[..., 0]
        bit = tf.equal(gl, max_gl) & (gl > thr_intensity) & mask
        posr = tf.cast(tf.where(bit), tf.int32)
        ts = tf.gather(times, posr[:, 0])
        rs = posr[:, 1]
        ys = posr[:, 2]
        xs = posr[:, 3]
        gs = tf.gather_nd(gl, posr)
        return tf.stack((ts, rs, ys, xs), axis=1), gs

    data = unmasked(data, mask)
    data = data.enumerate().shard(shard, 0).batch(batch)
    mask = K.constant(mask, tf.bool)
    gauss = K.constant(gauss)
    radius = K.constant(radius)
    thr_intensity = K.constant(thr_intensity)
    with trange((nt + shard - 1) // shard, disable=verbose == 0) as prog:
        pos, score = _find(data, mask, gauss, radius, thr_intensity, prog=prog)
    idx = tf.argsort(score)[::-1]
    return tf.gather(pos, idx).numpy(), tf.gather(score, idx).numpy()



import tensorflow.keras.backend as K
import tensorflow as tf
from tqdm import trange

from ..util.distribute import distributed, ReduceOp


def calc_std(data, mask, nt=None, verbose=1):

    @distributed(ReduceOp.CONCAT, ReduceOp.SUM, ReduceOp.SUM, ReduceOp.SUM)
    def _calc(imgs, mask):
        masked = tf.boolean_mask(imgs, mask, axis=1)
        avg_t = K.mean(masked, axis=1)
        sum_x = K.sum(imgs, axis=0)
        sumsq = K.sum(K.square(masked - avg_t[:, None]))
        nt = K.cast_to_floatx(K.shape(imgs)[0])
        return avg_t, sum_x, sumsq, nt

    mask = K.constant(mask, tf.bool)
    nx = K.cast_to_floatx(tf.math.count_nonzero(mask))
    with trange(nt, desc='Calc Stats', disable=verbose == 0) as prog:
        avg_t, sum_x, sumsq, nt = _calc(data, mask, prog=prog)
    avg_x = sum_x / nt
    avg_0 = K.mean(avg_t)
    avg_t -= avg_0
    avgxm = tf.boolean_mask(avg_x - avg_0, mask)
    std = K.sqrt(sumsq / nt / nx - K.mean(K.square(avgxm)))
    return avg_t.numpy(), avg_x.numpy(), std.numpy()



import tensorflow as tf
import numpy as np
import heapq as hq


def get_segment_index(gl, y, x, mask):
    # return get_segment_index_tf(gl, y, x, mask)
    return tf.numpy_function(
        get_segment_index_py, [gl, y, x, mask], tf.int32
    )


def get_segment_index_py(img, y0, x0, mask):

    def push(y, x, dy, dx):
        yn, xn = y + dy, x + dx
        if mask[yn, xn] and (img[yn, xn] <= img[y, x]):
            q.append([yn, xn, dy, dx])

    mask = np.pad(mask, [[0, 1], [0, 1]])

    pos = [[y0, x0]]
    q = []

    start = [
        [1, 0], [1, 1], [0, 1], [-1, 1], [-1, 0], [-1, -1], [0, -1], [1, -1]
    ]
    for dy, dx in start:
        push(y0, x0, dy, dx)

    while len(q) > 0:
        y, x, dy, dx = q[0]
        pos.append([y, x])
        q = q[1:]
        push(y, x, dy, dx)
        if dx == 0:
            push(y, x, dy, +1)
            push(y, x, dy, -1)
        elif dy == 0:
            push(y, x, +1, dx)
            push(y, x, -1, dx)

    return np.array(pos, np.int32)


def get_segment_index_tf(img, y0, x0, mask):

    def push(y, x, dy, dx):
        yn, xn = y+dy, x+dx
        if mask[yn, xn]:
            if img[yn, xn] <= img[y, x]:
                q.enqueue([yn, xn, dy, dx])

    mask = tf.pad(mask, [[0, 1], [0, 1]])
    pos = tf.TensorArray(tf.int32, size=0, dynamic_size=True)

    q = tf.queue.FIFOQueue(2000, [tf.int32]*4, [()]*4)
    i = tf.constant(0)
    pos = pos.write(i, [y0, x0])
    for dy in tf.constant([-1, 0, 1]):
        for dx in tf.constant([-1, 0, 1]):
            if tf.not_equal(dx, 0) & tf.not_equal(dy, 0):
                push(y0, x0, dy, dx)

    while q.size() > 0:
        y, x, dy, dx = q.dequeue()
        i += 1
        pos = pos.write(i, [y, x])
        push(y, x, dy, dx)
        if dx == 0:
            push(y, x, dy, +1)
            push(y, x, dy, -1)
        elif dy == 0:
            push(y, x, +1, dx)
            push(y, x, -1, dx)

    return pos.stack()



import tensorflow.keras.backend as K
import tensorflow as tf
import tensorflow_addons as tfa
import numpy as np

#from ..optimizer.prox_optimizer import ProxOptimizer as Optimizer
from ..optimizer.prox_nesterov import ProxNesterov as Optimizer
from ..optimizer.callback import Callback as OptCallback
from .variance import Variance


class BaseModel(tf.keras.Model):

    def __init__(self, footprint, spike, variance, **kwargs):
        super().__init__(**kwargs)
        self.footprint_penalty = footprint.penalty
        self.spike_penalty = spike.penalty
        self.la_set = footprint.set_l
        self.lu_set = spike.set_l
        self.variance = variance

    def set_penalty(self, la, lu, bx, bt):
        self.variance.set_baseline(bx, bt)
        nm = K.get_value(self.variance._nm)
        self.la_set(la / nm)
        self.lu_set(lu / nm)

    def compile(self, **kwargs):
        super().compile(
            optimizer=Optimizer(), loss=Loss(), **kwargs,
        )

    def call_common(self, val):
        variance = self.variance(val)
        loss = 0.5 * K.log(variance)
        footprint_penalty = self.footprint_penalty()
        spike_penalty = self.spike_penalty()
        me = loss + footprint_penalty + spike_penalty
        self.add_metric(K.sqrt(variance), 'sigma')
        self.add_metric(me, 'score')
        return loss, footprint_penalty, spike_penalty

    def fit_common(self, callback, log_dir=None, stage=None, callbacks=None,
                   steps_per_epoch=100, epochs=100, min_delta=1e-3, **kwargs):
        if callbacks is None:
            callbacks = []

        callbacks += [
            OptCallback(),
            tf.keras.callbacks.EarlyStopping(
                'score', min_delta=min_delta, patience=3,
                restore_best_weights=True, verbose=0,
            ),
        ]

        verbose = kwargs.pop('verbose', 1)
        if verbose != 0:
            callbacks += [
                tfa.callbacks.TQDMProgressBar(
                    leave_epoch_progress=False,
                    leave_overall_progress=True,
                    show_epoch_progress=False,
                    show_overall_progress=True,
                ),
            ]

        if log_dir is not None:
            callbacks += [
                callback(
                    log_dir=log_dir, stage=stage,
                    update_freq='batch', write_graph=False,
                ),
            ]

        dummy = tf.zeros((1, 1))
        data = tf.data.Dataset.from_tensor_slices((dummy, dummy)).repeat()
        super().fit(
            data,
            steps_per_epoch=steps_per_epoch,
            epochs=epochs,
            callbacks=callbacks,
            verbose=0,
            **kwargs,
        )


class Loss(tf.keras.losses.Loss):

    def call(self, y_true, y_pred):
        return y_pred



import tensorflow as tf
from tqdm import trange


def save_tfrecord(filebase, data, nt=None, verbose=1):
    with tf.io.TFRecordWriter(f'{filebase}.tfrecord') as writer:
        with trange(nt, desc='Save', disable=verbose == 0) as prog:
            for d in data:
                writer.write(tf.io.serialize_tensor(d).numpy())
                if prog is not None:
                    prog.update(1)


def load_tfrecord(filebase):
    data = tf.data.TFRecordDataset(f'{filebase}.tfrecord')
    data = data.map(lambda ex: tf.io.parse_tensor(ex, tf.float32))
    return data



from .base import Command, option, _option
from ..footprint.find import find_peak
from ..footprint.reduce import reduce_peak_idx
from ..util.dataset import unmasked
from ..util.tfrecord import load_tfrecord
from ..util.numpy import load_numpy, save_numpy
from ..util.pickle import load_pickle, save_pickle


class PeakCommand(Command):

    name = 'peak'
    description = 'Find peaks'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
        option('force', 'f'),
    ]

    def is_error(self, stage):
        return stage < 1

    def is_target(self, stage):
        return stage == 1

    def force_stage(self, stage):
        return 1

    def create(self, data, prev, curr, logs, gauss, radius, thr_intensity, thr_distance, shard):
        tfrecord = load_tfrecord(f'{data}-data')
        mask = load_numpy(f'{data}-mask')
        nt = load_pickle(f'{data}-stat')[1]
        batch = self.status.params['batch']
        verbose = self.status.params['pbar']
        pos, score = find_peak(
            tfrecord, mask, gauss, radius, thr_intensity, shard, batch, nt, verbose,
        )
        idx = reduce_peak_idx(pos, radius, thr_distance)
        pos = pos[idx]
        score = score[idx]
        save_pickle(f'{curr}-filter', (gauss, radius, shard))
        save_numpy(f'{curr}-peak', pos)
        save_numpy(f'{curr}-intensity', score)



import os

import tensorflow as tf
import numpy as np

from .base import Command, _option
from ..image.filter.gaussian import gaussian
from ..train.dynamics import SpikeToCalcium
from ..util.numpy import load_numpy
from ..util.pickle import load_pickle
from ..util.tiff import save_tiff
from ..util.csv import save_csv


class OutputCommand(Command):

    name = 'output'
    description = 'Output results'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
    ]

    def handle(self):
        self.set_job_dir()

        name = self.status.params['name']
        thr_out = self.status.params['thr-out']
        out_dir = os.path.join(self.application.job_dir, 'outs')
        out_base = os.path.join(out_dir, name)
        tf.io.gfile.makedirs(out_dir)

        history = self.status.history.get(name, ())
        stage = 3 * ((len(history) - 0) // 3) + 0
        data = self.status.find_saved(history[:1])
        spike = self.status.find_saved(history[:stage])
        segment = self.status.find_saved(history[:stage-1])
        data = os.path.join(self.work_dir, data, '000')
        spike = os.path.join(self.work_dir, spike, f'{stage:03d}')
        segment = os.path.join(self.work_dir, segment, f'{stage-1:03d}')

        mask0 = load_numpy(f'{data}-mask')
        h0, w0 = mask0.shape
        nx, nt, h, w, y0, x0, std = load_pickle(f'{data}-stat')
        mask = np.zeros((h, w), np.bool)
        mask[y0:y0+h0, x0:x0+w0] = mask0

        segment = load_numpy(f'{segment}-segment')
        spike = load_numpy(f'{spike}-spike')
        nk, nu = spike.shape

        out = np.zeros((nk, h, w), np.float32)
        out[:, mask] = segment
        out = gaussian(out, self.status.params['gauss']).numpy()
        out -= out.min(axis=(1, 2), keepdims=True)
        out /= out.max(axis=(1, 2), keepdims=True)
        out -= thr_out
        out[out < 0.0] = 0.0
        out /= (1 - thr_out)

        hz = self.status.params['hz']
        spike_to_calcium = SpikeToCalcium()
        spike_to_calcium.set_double_exp(*self.status.tau)
        calcium = spike_to_calcium(spike).numpy()
        pad = spike_to_calcium.pad
        time = (np.arange(nu) - pad) / hz
        spike /= spike.max(axis=1, keepdims=True)
        spike = np.concatenate([time[:, None], spike.T], axis=1)
        time = time[pad:]
        calcium = np.concatenate([time[:, None], calcium.T], axis=1)
        h = ','.join(['time'] + [f'id{k:04d}' for k in range(nk)])

        save_tiff(f'{out_base}-cell', out)
        save_csv(f'{out_base}-spike', spike, '%.3f', header=h)
        save_csv(f'{out_base}-calcium', calcium, '%.3f', header=h)



import sys
import os

import tensorflow as tf

from .base import Command, _option


class ConfigCommand(Command):

    name = 'config'
    description = 'Update parameter'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
        _option('name', None, ''),
        _option('pbar', None, ''),

        _option('imgs-file', None, ''),
        _option('mask-type', None, ''),

        _option('tau-rise', None, 'time constant of calicum raise (sec)'),
        _option('tau-fall', None, 'time constant of calcium fall (sec)'),
        _option('hz', None, 'sampling rate of data (1/sec)'),
        _option('tau-scale', None, ''),

        _option('gauss', 'g', 'size of gaussian filter (px)'),
        _option('radius-type', None, '{linear,log,manual}'),
        _option('radius', None, 'radius of cell (px)'),

        _option('min-intensity', None, ''),
        _option('min-distance', None, ''),
        _option('max-intensity', None, ''),
        _option('max-distance', None, ''),

        _option('thr-intensity', None, ''),
        _option('thr-distance', None, ''),
        _option('shard', None, ''),

        _option('la', 'a', 'penalty coefficient of footprint'),
        _option('lu', 'u', 'penalty coefficient of spike'),
        _option('bx', 'x', 'penalty coefficient of spatical baseline'),
        _option('bt', 't', 'penalty coefficient of temporal baseline'),
        _option('thr-firmness', None, ''),
        _option('thr-sim-area', None, ''),
        _option('thr-similarity', None, ''),

        _option('learning-rate', 'l', ''),
        _option('step', None, ''),
        _option('epoch', None, ''),
        _option('tol', None, ''),
        _option('batch', None, ''),
        _option('thr-out', None, ''),
    ]

    def handle(self):
        self.set_job_dir()
        self._update_parameter('name', 'default', str)
        self._update_parameter('pbar', 1, int)

        self._update_parameter('imgs-file', 'imgs.tif', str)
        self._update_parameter('mask-type', '0.pad', str)

        self._update_parameter('hz', 20.0)
        self._update_parameter('tau-rise', 0.08)
        self._update_parameter('tau-fall', 0.16)
        self._update_parameter('tau-scale', 6.0)

        self._update_parameter('gauss', 2.0)
        self._update_parameter('radius-type', 'log', str)
        self._update_parameter('radius', '2.0,40.0,13', float_tuple)

        self._update_parameter('min-intensity', 0.5)
        self._update_parameter('min-distance', 1.5)
        self._update_parameter('max-intensity', 1.0)
        self._update_parameter('max-distance', 2.0)
        self._update_parameter('thr-intensity', 0.7)
        self._update_parameter('thr-distance', 1.6)
        self._update_parameter('shard', 1, int)

        self._update_parameter('la', 10.0)
        self._update_parameter('lu', 100.0)
        self._update_parameter('bx', 0.5)
        self._update_parameter('bt', 0.5)
        self._update_parameter('thr-firmness', 0.1)
        self._update_parameter('thr-sim-area', 0.7)
        self._update_parameter('thr-similarity', 0.8)

        self._update_parameter('learning-rate', 0.01)
        self._update_parameter('step', 100, int)
        self._update_parameter('epoch', 100, int)
        self._update_parameter('tol', 1e-3)
        self._update_parameter('batch', 100, int)
        self._update_parameter('thr-out', 0.7)

        self.save_status()

        status = self.status.params
        for k, v in status.items():
            self.line('{}: {}'.format(k, v))

    def _update_parameter(self, name, default=None, dtype=float):
        status = self.status.params
        curr_val = status.get(name, None)
        val = self.option(name) or curr_val or default
        if val is None:
            self.line('missing: {name}', 'error')
            sys.exit(1)
        if val != curr_val:
            status[name] = dtype(val)


def float_tuple(val):
    return tuple(float(v) for v in val.split(','))



import tensorflow as tf

from .regularizer import get_prox


class ProxOptimizer(tf.keras.optimizers.Optimizer):

    _HAS_AGGREGATE_GRAD = True

    def __init__(self, learning_rate=1.0, name='Prox', **kwargs):
        super().__init__(name, **kwargs)
        self._set_hyper('learning_rate', kwargs.get('lr', learning_rate))

    def _resource_apply_dense(self, grad, var, apply_state=None):
        key = var.device, var.dtype.base_dtype
        coef = apply_state.get(key)
        lr = coef['lr_t']
        new_var = get_prox(var)(var - lr * grad, lr)
        return var.assign(new_var, use_locking=self._use_locking)

    def get_config(self):
        config = super().get_config()
        config.update(dict(
            learning_rate=self._serialize_hyperparameter('learning_rate'),
        ))
        return config



import tensorflow.keras.backend as K
import tensorflow as tf


def get_prox(var):
    if hasattr(var, 'regularizer') and hasattr(var.regularizer, 'prox'):
        return var.regularizer.prox
    else:
        return lambda x, _: x


def get_penalty(var):
    if hasattr(var, 'regularizer'):
        return var.regularizer(var)
    else:
        return 0.0


class ProxOp(tf.keras.layers.Layer):

    def call(self, x):
        return x

    def prox(self, y, eta):
        return y


class NonNegativeL1(ProxOp):

    def __init__(self):
        super().__init__()
        self.l = self.add_weight('l', (), trainable=False)

    def call(self, x):
        return self.l * K.sum(K.relu(x))

    def prox(self, y, eta):
        return K.relu(y - eta * self.l)


class MaxNormNonNegativeL1(ProxOp):

    def __init__(self, axis=-1):
        super().__init__()
        self.l = self.add_weight('l', (), trainable=False)
        self.axis = axis

    def call(self, x):
        x = K.relu(x)
        s = K.sum(x, axis=self.axis)
        m = K.max(x, axis=self.axis)
        cond = m > 0.0
        s = tf.boolean_mask(s, cond)
        m = tf.boolean_mask(m, cond)
        return self.l * K.sum(s / m)

    def prox(self, y, eta):
        y = K.relu(y)
        m = K.max(y, axis=self.axis, keepdims=True)
        return K.relu(tf.where(K.equal(y, m), y, y - eta * self.l / m))



import time

from .base import Command, _option


class RunCommand(Command):

    name = 'run'
    description = 'Execute all'
    help = '''
'''

    options = [
        _option('job-dir', 'j', 'target directory'),
        _option('goal'),
    ]

    def handle(self):
        start = time.time()

        self.set_job_dir()

        name = self.status.params['name']
        goal_stage = int(self.option('goal') or 30)

        stage = len(self.status.history.get(name, ()))
        for s in range(stage, goal_stage + 1):
            if s == 0:
                self.call('data')
            elif s == 1:
                self.call('peak')
            elif s == 2:
                self.call('segment')
            elif s % 3 == 0:
                self.call('spike')
            elif s % 3 == 1:
                self.call('footprint')
            elif s % 3 == 2:
                self.call('clean')

        self.call('output')

        s = time.time() - start
        h = int(s / 3600)
        s -= 3600 * h
        m = int(s / 60)
        s -= 60 * m
        self.line(f'{h}:{m}:{s}', 'comment')



def main():
    import os
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

    import tensorflow as tf
    tf.get_logger().setLevel('ERROR')

    # for debug
    #limit_gpu_memory = 7000
    limit_gpu_memory = None
    # tf.debugging.set_log_device_placement(True)
    # tf.config.experimental_run_functions_eagerly(True)
    # tf.autograph.set_verbosity(1, True)

    # limit GPU memory
    devs = tf.config.list_physical_devices('GPU')
    if limit_gpu_memory is not None:
        tf.config.set_logical_device_configuration(
            devs[0], [
                tf.config.LogicalDeviceConfiguration(
                    memory_limit=limit_gpu_memory,
                ),
            ],
        )
    else:
        for d in devs:
            tf.config.experimental.set_memory_growth(d, True)

    from .application import Application
    return Application().run()



import tensorflow.keras.backend as K
import tensorflow as tf
import numpy as np


def gaussian(imgs, r):
    sqrt_2pi = K.constant(np.sqrt(2.0 * np.pi))
    mr = tf.math.ceil(r)
    d = K.square(K.arange(-4.0 * mr, 4.0 * mr + 1.0, 1.0))
    r2 = K.square(r)
    o0 = K.exp(-0.5 * d / r2) / r / sqrt_2pi
    tmp = imgs[..., None]
    tmp = K.conv2d(tmp, tf.reshape(o0, (1, -1, 1, 1)), (1, 1), 'same')
    tmp = K.conv2d(tmp, tf.reshape(o0, (-1, 1, 1, 1)), (1, 1), 'same')
    return tmp[..., 0]



import tensorflow.keras.backend as K
import tensorflow as tf

from hotaru.train.dynamics import SpikeToCalcium, CalciumToSpike


class Variance(tf.keras.layers.Layer):

    def __init__(self, data, nk, nx, nt, name='Variance'):
        super().__init__(name=name, dtype=tf.float32)

        self._data = data
        self.nx = nx
        self.nt = nt

        self.spike_to_calcium = SpikeToCalcium(name='to_cal')
        self.calcium_to_spike = CalciumToSpike(name='to_spk')
        self._bx = 0.0
        self._bt = 0.0

        nz = max(nx, nt)
        self.nk = nk
        self._nn = self.add_weight('nn', (), trainable=False)
        self._nm = self.add_weight('nm', (), trainable=False)
        self._dat = self.add_weight('dat', (nk, nz), trainable=False)
        self._cov = self.add_weight('cov', (nk, nk), trainable=False)
        self._out = self.add_weight('out', (nk, nk), trainable=False)

    def set_double_exp(self, *tau):
        self.spike_to_calcium.set_double_exp(*tau)
        self.calcium_to_spike.set_double_exp(*tau)
        self.nu = self.nt + self.calcium_to_spike.pad

    def set_baseline(self, bx, bt):
        nxf = K.constant(self.nx)
        ntf = K.constant(self.nt)
        nn = nxf * ntf
        nm = nn
        if bx > 0.0:
            nm += nxf
        if bt > 0.0:
            nm += ntf
        self._bx = bx
        self._bt = bt
        K.set_value(self._nn, nn)
        K.set_value(self._nm, nm)

    def call(self, xdat):
        nk, nx = tf.shape(xdat)[0], tf.shape(xdat)[1]
        xcov = tf.matmul(xdat, xdat, False, True)
        xsum = K.sum(xdat, axis=1)
        xout = xsum[:, None] * xsum / tf.cast(nx, tf.float32)

        ydat = tf.slice(self._dat, [0, 0], [nk, nx])
        ycov = tf.slice(self._cov, [0, 0], [nk, nk])
        yout = tf.slice(self._out, [0, 0], [nk, nk])

        variance = (
            self._nn
            + K.sum(ydat * xdat)
            + K.sum(ycov * xcov)
            + K.sum(yout * xout)
        ) / self._nm
        return variance

    def _cache(self, mode, yval, dat):
        if mode == 0:
            ny, bx, by = self.nx, self._bt, self._bx
        else:
            ny, bx, by = self.nt, self._bx, self._bt
        max_nk, nmax = tf.shape(self._dat)

        ycov = tf.matmul(yval, yval, False, True)
        ysum = K.sum(yval, axis=1)
        yout = ysum[:, None] * ysum / ny
        cx = 1.0 - K.square(bx)
        cy = 1.0 - K.square(by)
        dat = -2.0 * dat
        cov = ycov - cx * yout
        out = yout - cy * ycov

        lipschitz = K.max(tf.linalg.eigvalsh(cov)) / self._nm
        gsum = K.sum(self.spike_to_calcium.kernel)
        self.lipschitz_a = lipschitz.numpy()
        self.lipschitz_u = (lipschitz * gsum).numpy()

        nk, nx = tf.shape(dat)
        dk = max_nk - nk
        dx = nmax - nx
        dat = tf.pad(dat, [[0, dk], [0, dx]])
        cov = tf.pad(cov, [[0, dk], [0, dk]])
        out = tf.pad(out, [[0, dk], [0, dk]])

        K.set_value(self._dat, dat)
        K.set_value(self._cov, cov)
        K.set_value(self._out, out)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['hotaru',
 'hotaru.console',
 'hotaru.footprint',
 'hotaru.image',
 'hotaru.image.filter',
 'hotaru.optimizer',
 'hotaru.train',
 'hotaru.util']

package_data = \
{'': ['*']}

install_requires = \
['cleo>=0.8,<0.9',
 'matplotlib>=3.5,<4.0',
 'scipy>=1.8,<2.0',
 'tensorflow-addons>=0.16,<0.17',
 'tensorflow>=2.8,<3.0',
 'tifffile>=2022,<2023',
 'tqdm>=4.63,<5.0']

entry_points = \
{'console_scripts': ['hotaru = hotaru.console:main']}

setup_kwargs = {
    'name': 'hotaru',
    'version': '3.4.1',
    'description': 'High performance Optimizer to extract spike Timing And cell location from calcium imaging data via lineaR impUlse',
    'long_description': '# HOTARU\n\nHigh performance Optimizer to extract spike Timing And cell location from calcium imaging data via lineaR impUlse\n\n### Author\nTAKEKAWA Takashi <takekawa@tk2lab.org>\n\n### Reference\n- Takekawa., T, et. al.,\n  Automatic sorting system for large scale calcium imaging data, bioRxiv (2017).\n  https://doi.org/10.1101/215145\n\n\n## Install\n\n### Require\n- python >= 3.8\n- tensorflow >= 2.6.1\n\n### Install Procedure (using venv)\n- Create venv environment for hotaru\n  - `python3.x -m venv hotaru`\n- Activate hotaru environment\n  - `source hotaru/bin/activate`\n- Install hotaru\n  - `pip install hotaru`\n\n\n## Usage\n\n### Apply Method\n- (in hotaru venv)\n- `mkdir work`\n- `cd work`\n- `cp somewhere/TARGET.tif imgs.tif`\n- `hotaru config`\n- `hotaru run`\n- (see outs directory)\n\n### Config Option\n- Set sampling rate of movie  \n  `hotaru config --hz 20.0`\n- Set mask file (tif or npy)  \n  `hotaru config --mask-type mask.tif`\n- Set calcium dynamics  \n  `hotaru config --tau-rise 0.08 --tau-fall 0.16`\n- Set cell size candidate  \n  `hotaru config --radius-type log --radius "2.0,40.0,13"`  \n  `hotaru config --radius-type linear --radius "2.0,11.0,10"`  \n  `hotaru config --radius-type manual --radius "2,3,4,5,6,7,8,9,10"`    \n  \n### Check Resutls\n- (in hotaru venv and in work dir)\n- `tensorboard --logidr logs`\n- open in web browser `http://localhost:6006`\n',
    'author': 'TAKEKAWA Takashi',
    'author_email': 'takekawa@tk2lab.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tk2lab/HOTARU',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import tensorflow.keras.backend as K


def neighbor(imgs):
    fil = K.constant([[1, 1, 1], [1, 0, 1], [1, 1, 1]]) / 8.0
    fil = K.reshape(fil, [3, 3, 1, 1])
    return K.conv2d(
        imgs[..., None], fil, [1, 1, 1, 1], padding='SAME',
    )[..., 0]
